VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "BSPTree"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
Option Explicit

Private m_Triangles As Collection

Public SplittingTriangle As SimplePolygon
Public ChildPlus As BSPTree
Public ChildMinus As BSPTree
' Add this triangle to the appropriate child.
Public Sub AddTriangle(ByVal triangle As SimplePolygon)
Const PRECISION As Single = 0.001
Dim Vx As Single
Dim Vy As Single
Dim Vz As Single
Dim distance(1 To 3) As Single
Dim i As Integer
Dim on_plus_side As Integer
Dim on_minus_side As Integer
Dim A As Integer
Dim B As Integer
Dim x1 As Single
Dim y1 As Single
Dim z1 As Single
Dim x2 As Single
Dim y2 As Single
Dim z2 As Single

    ' Make sure the triangle has three vertices.
    If triangle.PointX.Count < 3 Then Exit Sub

    ' See on which side the triangle's points belong.
    For i = 1 To 3
        ' Get the signed distance from the point to the plane.
        Vx = triangle.PointX(i) - SplittingTriangle.PointX(1)
        Vy = triangle.PointY(i) - SplittingTriangle.PointY(1)
        Vz = triangle.PointZ(i) - SplittingTriangle.PointZ(1)
        distance(i) = _
            Vx * SplittingTriangle.Nx + _
            Vy * SplittingTriangle.Ny + _
            Vz * SplittingTriangle.Nz
        If distance(i) > PRECISION Then
            on_plus_side = on_plus_side + 1
        ElseIf distance(i) < -PRECISION Then
            on_minus_side = on_minus_side + 1
        End If
    Next i

    ' See if all three points are on the same side.
    If (on_plus_side > 0) And (on_minus_side = 0) Then
        ' They're all on the plus side.
        ' Add the triangle to the plus child.
        ' Create the child if necessary.
        If ChildPlus Is Nothing Then
            ' Create the child.
            Set ChildPlus = New BSPTree
            ChildPlus.SetSplittingTriangle triangle
        Else
            ' The child exists. Add the triangle to it.
            ChildPlus.AddTriangle triangle
        End If
    ElseIf (on_plus_side = 0) And (on_minus_side > 0) Then
        ' They're all on the minus side.
        ' Add the triangle to the minus child.
        ' Create the child if necessary.
        If ChildMinus Is Nothing Then
            ' Create the child.
            Set ChildMinus = New BSPTree
            ChildMinus.SetSplittingTriangle triangle
        Else
            ' The child exists. Add the triangle to it.
            ChildMinus.AddTriangle triangle
        End If
    ElseIf (on_plus_side = 0) And (on_minus_side = 0) Then
        ' They're all on the splitting plane.
        ' Add it to the list of triangles we must draw.
        m_Triangles.Add triangle
    Else
        ' The triangle is split.
        ' See if a point is on the splitting plane.
        If on_plus_side + on_minus_side < 3 Then
            ' A point is on the splitting plane.
            ' Break the triangle into two.
            ' Find the point on the plane.
            For i = 1 To 3
                If (distance(i) <= PRECISION) And (distance(i) >= -PRECISION) Then Exit For
            Next i

            ' Find the other two points.
            A = i + 1
            If A > 3 Then A = 1
            B = A + 1
            If B > 3 Then B = 1

            ' See where line AB intersects the plane.
            IntersectLineWithPlane _
                triangle.PointX(A), _
                triangle.PointY(A), _
                triangle.PointZ(A), _
                triangle.PointX(B), _
                triangle.PointY(B), _
                triangle.PointZ(B), _
                x1, y1, z1

            ' Make the new triangles.
            AddTriangle NewTriangle( _
                x1, y1, z1, _
                triangle.PointX(A), triangle.PointY(A), triangle.PointZ(A), _
                triangle.PointX(i), triangle.PointY(i), triangle.PointZ(i), _
                triangle)
            AddTriangle NewTriangle( _
                x1, y1, z1, _
                triangle.PointX(B), triangle.PointY(B), triangle.PointZ(B), _
                triangle.PointX(i), triangle.PointY(i), triangle.PointZ(i), _
                triangle)
        Else ' End if there's a point on the splitting plane.
            ' No points are on the splitting plane.
            ' Find the point that's alone.
            If on_plus_side = 1 Then
                ' There's one point on the plus side.
                For i = 1 To 3
                    If distance(i) > PRECISION Then Exit For
                Next i
            Else
                ' There's one point on the minus side.
                For i = 1 To 3
                    If distance(i) < -PRECISION Then Exit For
                Next i
            End If

            ' Find the other two points.
            A = i + 1
            If A > 3 Then A = 1
            B = A + 1
            If B > 3 Then B = 1

            ' Find the points where the triangle
            ' intersects the splitting plane.
            IntersectLineWithPlane _
                triangle.PointX(i), _
                triangle.PointY(i), _
                triangle.PointZ(i), _
                triangle.PointX(A), _
                triangle.PointY(A), _
                triangle.PointZ(A), _
                x1, y1, z1
            IntersectLineWithPlane _
                triangle.PointX(i), _
                triangle.PointY(i), _
                triangle.PointZ(i), _
                triangle.PointX(B), _
                triangle.PointY(B), _
                triangle.PointZ(B), _
                x2, y2, z2

            ' Break the triangle into three.
            AddTriangle NewTriangle( _
                x1, y1, z1, _
                x2, y2, z2, _
                triangle.PointX(i), triangle.PointY(i), triangle.PointZ(i), _
                triangle)
            AddTriangle NewTriangle( _
                x1, y1, z1, _
                x2, y2, z2, _
                triangle.PointX(A), triangle.PointY(A), triangle.PointZ(A), _
                triangle)
            AddTriangle NewTriangle( _
                x2, y2, z2, _
                triangle.PointX(A), triangle.PointY(A), triangle.PointZ(A), _
                triangle.PointX(B), triangle.PointY(B), triangle.PointZ(B), _
                triangle)
        End If  ' End the triangle has no point
                ' on the splitting plane.
    End If ' End the triangle is split.
End Sub

' Draw the BSP tree with the given center of projection.
Public Sub DrawBSPTree(ByVal X As Single, ByVal Y As Single, ByVal Z As Single, ByVal pic As PictureBox)
Dim Vx As Single
Dim Vy As Single
Dim Vz As Single
Dim distance As Single
Dim cosine As Single
Dim pgon As SimplePolygon

    ' Get the signed distance from the
    ' center of projection to the plane.
    Vx = X - SplittingTriangle.PointX(1)
    Vy = Y - SplittingTriangle.PointY(1)
    Vz = Z - SplittingTriangle.PointZ(1)
    distance = _
        Vx * SplittingTriangle.Nx + _
        Vy * SplittingTriangle.Ny + _
        Vz * SplittingTriangle.Nz

    If distance > 0 Then
        ' The center of project is on the plus side.
        ' Draw the minus side first, then our
        ' splitting triangle, then the plus side.
        If Not (ChildMinus Is Nothing) Then
            ChildMinus.DrawBSPTree X, Y, Z, pic
        End If

        For Each pgon In m_Triangles
            pgon.DrawPolygon pic
        Next pgon

        If Not (ChildPlus Is Nothing) Then
            ChildPlus.DrawBSPTree X, Y, Z, pic
        End If
    Else
        ' The center of project is on the minus side.
        ' Draw the plus side first, then our
        ' splitting triangle, then the minus side.
        If Not (ChildPlus Is Nothing) Then
            ChildPlus.DrawBSPTree X, Y, Z, pic
        End If

        For Each pgon In m_Triangles
            pgon.DrawPolygon pic
        Next pgon

        If Not (ChildMinus Is Nothing) Then
            ChildMinus.DrawBSPTree X, Y, Z, pic
        End If
    End If
End Sub

' Find the point (X, Y, Z) where the line between
' (X1, Y1, Z1) and (X2, Y2, Z2) intersects the
' splitting plane.
Private Sub IntersectLineWithPlane(ByVal x1 As Single, ByVal y1 As Single, ByVal z1 As Single, ByVal x2 As Single, ByVal y2 As Single, ByVal z2 As Single, ByRef X As Single, ByRef Y As Single, ByRef Z As Single)
Dim Vx As Double
Dim Vy As Double
Dim Vz As Double
Dim numer As Double
Dim denom As Double
Dim t As Double

    Vx = x2 - x1
    Vy = y2 - y1
    Vz = z2 - z1
    numer = _
        SplittingTriangle.Nx * (x1 - SplittingTriangle.PointX(1)) + _
        SplittingTriangle.Ny * (y1 - SplittingTriangle.PointY(1)) + _
        SplittingTriangle.Nz * (z1 - SplittingTriangle.PointZ(1))
    denom = _
        SplittingTriangle.Nx * Vx + _
        SplittingTriangle.Ny * Vy + _
        SplittingTriangle.Nz * Vz
    t = -numer / denom

    X = x1 + t * Vx
    Y = y1 + t * Vy
    Z = z1 + t * Vz
End Sub
' Make a new triangle.
Private Function NewTriangle(ByVal x1 As Single, ByVal y1 As Single, ByVal z1 As Single, ByVal x2 As Single, ByVal y2 As Single, ByVal z2 As Single, ByVal X3 As Single, ByVal Y3 As Single, ByVal Z3 As Single, ByVal pgon As SimplePolygon) As SimplePolygon
Dim tri As SimplePolygon

    Set tri = New SimplePolygon
    With tri
        .AddPoint x1, y1, z1
        .AddPoint x2, y2, z2
        .AddPoint X3, Y3, Z3
        .FillColor = pgon.FillColor
        .ForeColor = pgon.ForeColor
    End With
    Set NewTriangle = tri
End Function

' Return the number of triangles in the BSP tree.
Public Property Get NumTriangles() As Integer
Dim num As Integer

    num = 1
    If Not (ChildPlus Is Nothing) Then num = num + ChildPlus.NumTriangles
    If Not (ChildMinus Is Nothing) Then num = num + ChildMinus.NumTriangles
    NumTriangles = num
End Property

' Set the node's splitting triangle.
Public Sub SetSplittingTriangle(ByVal triangle As SimplePolygon)
    ' Save the splitting triangle.
    Set SplittingTriangle = triangle
    triangle.Finish

    ' Add this triangle to our list of triangles.
    m_Triangles.Add triangle
End Sub

Private Sub Class_Initialize()
    Set m_Triangles = New Collection
End Sub


